# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tts.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\ttts.proto\x12\x0c\x63hat_bot.tts\"K\n\x10LoadModelRequest\x12\x0f\n\x07tts_tag\x18\x01 \x01(\t\x12\x11\n\tis_reload\x18\x02 \x01(\x08\x12\x13\n\x0bjson_kwargs\x18\x03 \x01(\t\"\x13\n\x11LoadModelResponse\":\n\x11SynthesizeRequest\x12\x10\n\x08tts_text\x18\x01 \x01(\t\x12\x13\n\x0bjson_kwargs\x18\x02 \x01(\t\"\'\n\x12SynthesizeResponse\x12\x11\n\ttts_audio\x18\x01 \x01(\x0c\"\x12\n\x10GetVoicesRequest\"#\n\x11GetVoicesResponse\x12\x0e\n\x06voices\x18\x01 \x03(\t\" \n\x0fSetVoiceRequest\x12\r\n\x05voice\x18\x01 \x01(\t\"\x12\n\x10SetVoiceResponse\"\x16\n\x14GetStreamInfoRequest\"\\\n\x14GetStreamInfoReponse\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\x05\x12\x10\n\x08\x63hannels\x18\x02 \x01(\x05\x12\x0c\n\x04rate\x18\x03 \x01(\x05\x12\x14\n\x0csample_width\x18\x04 \x01(\x05\x32\xa4\x03\n\x03TTS\x12N\n\tLoadModel\x12\x1e.chat_bot.tts.LoadModelRequest\x1a\x1f.chat_bot.tts.LoadModelResponse\"\x00\x12N\n\tGetVoices\x12\x1e.chat_bot.tts.GetVoicesRequest\x1a\x1f.chat_bot.tts.GetVoicesResponse\"\x00\x12K\n\x08SetVoice\x12\x1d.chat_bot.tts.SetVoiceRequest\x1a\x1e.chat_bot.tts.SetVoiceResponse\"\x00\x12Y\n\rGetStreamInfo\x12\".chat_bot.tts.GetStreamInfoRequest\x1a\".chat_bot.tts.GetStreamInfoReponse\"\x00\x12U\n\x0cSynthesizeUS\x12\x1f.chat_bot.tts.SynthesizeRequest\x1a .chat_bot.tts.SynthesizeResponse\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tts_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_LOADMODELREQUEST']._serialized_start=27
  _globals['_LOADMODELREQUEST']._serialized_end=102
  _globals['_LOADMODELRESPONSE']._serialized_start=104
  _globals['_LOADMODELRESPONSE']._serialized_end=123
  _globals['_SYNTHESIZEREQUEST']._serialized_start=125
  _globals['_SYNTHESIZEREQUEST']._serialized_end=183
  _globals['_SYNTHESIZERESPONSE']._serialized_start=185
  _globals['_SYNTHESIZERESPONSE']._serialized_end=224
  _globals['_GETVOICESREQUEST']._serialized_start=226
  _globals['_GETVOICESREQUEST']._serialized_end=244
  _globals['_GETVOICESRESPONSE']._serialized_start=246
  _globals['_GETVOICESRESPONSE']._serialized_end=281
  _globals['_SETVOICEREQUEST']._serialized_start=283
  _globals['_SETVOICEREQUEST']._serialized_end=315
  _globals['_SETVOICERESPONSE']._serialized_start=317
  _globals['_SETVOICERESPONSE']._serialized_end=335
  _globals['_GETSTREAMINFOREQUEST']._serialized_start=337
  _globals['_GETSTREAMINFOREQUEST']._serialized_end=359
  _globals['_GETSTREAMINFOREPONSE']._serialized_start=361
  _globals['_GETSTREAMINFOREPONSE']._serialized_end=453
  _globals['_TTS']._serialized_start=456
  _globals['_TTS']._serialized_end=876
# @@protoc_insertion_point(module_scope)
